DEFAULT_AUTH_PATH = "/api/rest/v1/auth"
DEFAULT_BASE_PATH = "/api/json/v1"

DEFAULT_TIMEOUT = 30000
DEFAULT_AUTH_HEADER = "X-Cassandra-Token"
DEFAULT_KEYSPACE_NAME = "default_keyspace"
DEFAULT_REGION = "us-east1"

DEFAULT_DEV_OPS_PATH_PREFIX = "/v2"
DEFAULT_DEV_OPS_URL = "api.astra.datastax.com"
